#include "ConsensusData.hh"

namespace libSlice
{

/*! @file
 *  @brief Defines implementation for ConsensusData
 */

//! Constructor stores the calculations values
/*! 
 *  @param consensusResults Results of getConsensus with ambiguity off
 *  @param ambiguityCode Ambiguity code of consensus
 *  @param qualityClass Quality class of slice
 *  @param doAmbiguity Records if ambiguity code was used or not
 */
ConsensusData::ConsensusData(libSlice_Consensus * consensusResults,
                             char ambiguityCode,
                             char qualityClass,
                             int doAmbiguity)
{
  memcpy(&m_consensusResults, consensusResults, sizeof(m_consensusResults));
  m_qualityClass = qualityClass;
  m_ambiguityCode = ambiguityCode;
  m_doAmbiguity = doAmbiguity;
}

//! Empty Deconstructor 
ConsensusData::~ConsensusData()
{
}

//! Returns the consensus (which may be an ambiguity code)
char ConsensusData::getConsensus() const
{
  if (m_doAmbiguity) return getAmbiguityCode();
  return getConsensusBase();
}

//! Gets the consensus as an ambiguity code (which may be a single base)
char ConsensusData::getAmbiguityCode() const
{
  return m_ambiguityCode;
}

//! Return a single base consensus
char ConsensusData::getConsensusBase() const
{
  return m_consensusResults.consensus;
}

//! Get the raw bit flags for which of the 5 bases are in the ambiguity code
char ConsensusData::getAmbiguityFlags() const
{
  return m_consensusResults.ambiguityFlags;
}

//! Gets the quality class of the slice
int ConsensusData::getQualityClass() const
{
  return (int) m_qualityClass;
}

//! Gets the consensus quality value of the slice
int ConsensusData::getQualityValue() const
{
  return m_consensusResults.qvConsensus;
}

//! Gets the consensus quality value of the 5 different bases 
/*! The consensus quality value of each base is a representation of the 
 *  error probability that the given base is the true consensus.
 *
 *  @param base Base in {'A','C','G','T','-'} 
 *  @retval Consensus quality value of the specified base
 */
int ConsensusData::getQualityValue(char base) const
{
  int retval;

  switch (base)
  {
    case 'A': retval = m_consensusResults.qvA;   break;
    case 'C': retval = m_consensusResults.qvC;   break;
    case 'G': retval = m_consensusResults.qvG;   break;
    case 'T': retval = m_consensusResults.qvT;   break;
    case '-': retval = m_consensusResults.qvGap; break;

    default: retval = m_consensusResults.qvConsensus; break;
  };

  return retval;
}
  
//! Tests if the given base is part of the ambiguity code
/*! This allows for tests to check the presense of any of the bases in the 
 *  ambiguity code, including testing if '-' should be a part of the code
 *  in the extended ambiguity code set.
 *  
 *  @param base Base in {'A','C','G','T','-'} 
 *  @retval Bool indicating if base should be included
 */
int ConsensusData::testAmbiguityCode(char base) const
{
  int retval = 0;

  switch (base)
  {
    case 'A': 
      retval = m_consensusResults.ambiguityFlags & AMBIGUITY_FLAGBIT_A;   
      break;
    case 'C': 
      retval = m_consensusResults.ambiguityFlags & AMBIGUITY_FLAGBIT_C;   
      break;
    case 'G': 
      retval = m_consensusResults.ambiguityFlags & AMBIGUITY_FLAGBIT_G;   
      break;
    case 'T': 
      retval = m_consensusResults.ambiguityFlags & AMBIGUITY_FLAGBIT_T;   
      break;
    case '-': 
      retval = m_consensusResults.ambiguityFlags & AMBIGUITY_FLAGBIT_GAP;   
      break;

    default:
      retval = 0;
  }

  return (retval != 0);
}

//! Gets the probability of error that the given base is the consensus
/*! 
 *  @param base Base in {'A','C','G','T','-'} 
 *  @retval Error probability that the given base is the consensus
 */
long double ConsensusData::getProbability(char base) const
{
  long double retval;

  switch (base)
  {
    case 'A': retval = m_consensusResults.cpA;   break;
    case 'C': retval = m_consensusResults.cpC;   break;
    case 'G': retval = m_consensusResults.cpG;   break;
    case 'T': retval = m_consensusResults.cpT;   break;
    case '-': retval = m_consensusResults.cpGap; break;

    default: retval = getProbability(getConsensusBase());
  };

  return retval;
}

}
