#include "Exceptions.hh"
#include "ReadData.hh"

#include <stdlib.h> // for atol

/*! @file
 *  @brief Defines implementation for ReadData and ReadManager
 */

namespace libSlice
{
  using std::string;

//! Simple constructor copies the values and parses the direction
/*! 
 *  @param id Id for the read as referenced by slice information
 *  @param seqname Internal name of the sequence [optional]
 *  @param dir Direction of read [forward if null]
 *  @param chem Chemistry of read [optional]
 *
 *  @throws Exception if id is NULL
 *  @see parseDirection
 */
ReadData::ReadData(const char * id, 
                   const char * seqname,
                   const char * dir, 
                   const char * chem)
{
  if (!id)
  {
    throw Exception("Cannot have NULL read id", "ReadData::ReadData");
  }

  m_id = id;

  m_dir     = (dir)     ? dir     : "0";
  m_chem    = (chem)    ? chem    : "";
  m_seqname = (seqname) ? seqname : "";

  m_clrleft  = m_seqleft = 0;
  m_clrright = m_seqright = 0;

  parseDirection();
}

//! Adds CLR data from Slice XML
/*! m_seqleft is atol(left), m_seqright is atol(right) <br>
 *
 *  m_clrleft is m_seqleft if !rc, m_seqright if rc <br>
 *  m_clrright is m_seqright if !rc, m_seqleft if rc <br>
 *
 *  @throws Exception if m_clrright < m_clrleft
 */
void ReadData::addClrData(const char * left,
                          const char * right)
{
  setSeqData((left)  ? atol(left)  : 0,
             (right) ? atol(right) : 0);
}

void ReadData::setSeqData(long int seqleft,
                          long int seqright)
{
  m_seqleft  = seqleft;
  m_seqright = seqright;

  m_clrleft  = (m_rc) ? m_seqright : m_seqleft;
  m_clrright = (m_rc) ? m_seqleft  : m_seqright;

  if (m_clrright < m_clrleft)
  {
    throw Exception("Invalid Clrdata for id " + m_id,
                        "ReadData::addClrData");
  }
}

void ReadData::setClrData(long int clrleft,
                          long int clrright)
{
  if (m_rc)
  {
    setSeqData(clrright, clrleft);
  }
  else
  {
    setSeqData(clrleft, clrright);

  }
}


//! Parses the direction string into the number 0 or 1
/*! The string "0" or "Forward" means that it is NOT reverse compliment
 *  The string "1" or "Reversed" means that it is reverse compliment
 *
 *  @throws Exception if direction is not understood
 */
void ReadData::parseDirection()
{
  if (m_dir == "0" || m_dir[0] == 'F')
  {
    m_rc = 0;
  }
  else if (m_dir == "1" || m_dir[0] == 'R')
  {
    m_rc = 1;
  }
  else
  {
    throw Exception("Read Direction is unknown in " + m_dir,
                        "ReadData::ReadData");
  }
}

//! Simple getter to get if read is reverse compliment
/*!
 *  @returns (char) 0 or (char) 1 if it is reverse compliment
 */
char ReadData::isReverseCompliment() const
{
  return m_rc;
}

//! Prints out all of the read information in XML format
/*! Only prints clr range or trimmed sequences if provided
 *
 *  @param out Stream to write to
 *  @param prefix Prefix for each line
 *
 */
void ReadData::printXML(ostream & out, const string & prefix) const
{
  out << prefix 
      << "<Read Id=\""    << m_id
      << "\" Seqname=\""  << m_seqname
      << "\" Dir=\""      << m_dir
      << "\" Chem=\""     << m_chem 
      << "\">" << endl;

  if (m_seqright)
  {
    out << prefix 
        << "  <Clr Left=\"" << m_seqleft
        << "\" Right=\""    << m_seqright
        << "\" />" << endl;
  }

  if (m_leftnuc.length())
  {
    out << prefix << "  <TrimmedSeq position=\"left\">"            << endl
        << prefix << "    <Nuc>"     << m_leftnuc  << "</Nuc>"     << endl
        << prefix << "    <Qualval>" << m_leftqual << "</Qualval>" << endl
        << prefix << "  </TrimmedSeq>"                             << endl;
  }

  if (m_rightnuc.length())
  {
    out << prefix << "  <TrimmedSeq position=\"right\">"            << endl
        << prefix << "    <Nuc>"     << m_rightnuc  << "</Nuc>"     << endl
        << prefix << "    <Qualval>" << m_rightqual << "</Qualval>" << endl
        << prefix << "  </TrimmedSeq>"                              << endl;
  }

  out << prefix << "</Read>" << endl;
}

//! Reverse compliments the read
/*! Reverse the direction, and swap the clear range attributes
 */
void ReadData::reverseCompliment()
{
  if (m_dir == "Forward")
  {
    m_dir = "Reversed";
  }
  else if (m_dir == "0")
  {
    m_dir = "1";
  }
  else if (m_dir == "Reversed")
  {
    m_dir = "Forward";
  }
  else if (m_dir == "1")
  {
    m_dir = "0";
  }
 
  long swap = m_seqleft;
  m_seqleft = m_seqright;
  m_seqright = swap;
}

//! Resets the id of a read.
/*! Use readManager::resetReadId or flatten slice will fail to find the
 *  corresponding id
 *
 *  Note: If you reset the id, you will become responsible for "fixing"
 *        the slices that reference it.
 *
 *  @param id New id for slice
 */
void ReadData::resetId(const string & id)
{
  m_id = id;
}



//! Gets the id of a read
const string & ReadData::getId() const
{
  return m_id;
}

//! Gets the seqname of a read
const string & ReadData::getSeqname() const
{
  return m_seqname;
}

//! Gets the chem of a read
const string & ReadData::getChem() const
{
  return m_chem;
}

//! Gets the seq left from a read
long ReadData::getSeqleft() const
{
  return m_seqleft;
}

//! Gets the seq right from a read
long ReadData::getSeqright() const
{
  return m_seqright;
}

//! Gets the clr left from a read
long ReadData::getClrleft() const
{
  return m_clrleft;
}

//! Gets the clr right from a read
long ReadData::getClrright() const
{
  return m_clrright;
}

//! Adds TrimmedSeq Nuc Data
void ReadData::addNucData(const char * buffer, bool isLeft)
{
  if (buffer)
  {
    if (isLeft)
    {
      m_leftnuc += buffer;
    }
    else
    {
      m_rightnuc += buffer;
    }
  }
}

//! Adds TrimmedSeq QualData
void ReadData::addQualData(const char * buffer, bool isLeft)
{
  if (buffer)
  {
    if (isLeft)
    {
      m_leftqual += buffer;
    }
    else
    {
      m_rightqual += buffer;
    }
  }
}

//! Get the raw left or right nuc data without any reverse compliment
const string & ReadData::getRawNucData(bool getLeft) const
{
  if (getLeft) return m_leftnuc;
  return m_rightnuc;
}


//! Get the raw left or right qual data without any reverse compliment
const string & ReadData::getRawQualData(bool getLeft) const
{
  if (getLeft) return m_leftqual;
  return m_rightqual;
}

//! Validates the TrimmedSequence Data
void ReadData::validateSeqdata() const
{
  //TODO: Implement this

}


//! Gets the left or right Nuc Data
/*! Prepends "?" for elements that occur outside of the clear range
 *  if they were not provided as TrimmedSeq data
 */
const string ReadData::getNucData(bool getLeft) const
{
  validateSeqdata();

  string retval;

  if (getLeft)
  {
    retval = m_leftnuc;

    if (!retval.size())
    {
      for (long i = 0; i < m_clrleft-1; i++)
      {
        retval += "?";
      }
    }
  }
  else
  {
    retval = m_rightnuc;
    
    // Can't pad here, because length is unknown
  }

  return retval;
}

//! Gets the left or right Qual Data
/*! Prepends "0" for elements that occur outside of the clear range
 *  if they were not provided as TrimmedSeq data
 */
const string ReadData::getQualData(bool getLeft) const
{
  validateSeqdata();

  string retval;

  if (getLeft)
  {
    retval = m_leftqual;

    if (!retval.size())
    {
      bool isFirst = true;
      for (long i = 0; i < m_clrleft-1; i++)
      {
        if (!isFirst) retval += " ";
        retval += "00";
        isFirst = false;
      }
    }
  }
  else
  {
    retval = m_rightqual;
    
    // Can't pad here, because length is unknown
  }

  return retval;
}


//! Truncates the TrimmedSeq nucs and quals
void ReadData::flushTrimmedSeq()
{
  m_leftnuc   = "";
  m_leftqual  = "";

  m_rightnuc  = "";
  m_rightqual = "";
}

//! Truncates the TrimmedSeq nucs and quals on a single side
void ReadData::flushTrimmedSeqSide(bool flushLeft)
{
  if (flushLeft)
  {
    m_leftnuc = "";
    m_leftqual = "";
  }
  else
  {
    m_rightnuc = "";
    m_rightqual = "";
  }
}


////////////// ReadManager Implementation

//! Simple Constructor
ReadManager::ReadManager()
{
  m_readIterator = m_readMap.end();
}

//! Deletes all of the reads contained in the map
ReadManager::~ReadManager()
{
  ReadMap::iterator i;

  for (i =  m_readMap.begin();
       i != m_readMap.end();
       i++)
  {
    delete i->second;
  } 
}

//! Adds a named read and direction and chemistry
/*! Creates the new read, and stores it in the lookup map
 *  @see ReadData
 *  @see insertRead
 *  
 *  @retval Pointer to the ReadData object just created
 */
ReadData * ReadManager::addReadData(const char * id,
                                    const char * seqname,
                                    const char * dir,
                                    const char * chem)
{
  ReadData * newRead = new ReadData(id, seqname, dir, chem);

  insertRead(newRead);

  return newRead;
}

//! Iterates through the list of reads
ReadData * ReadManager::getNextRead()
{
  ReadData * retval = NULL;
  
  if (m_readIterator != m_readMap.end())
  {
    retval = m_readIterator->second;
    m_readIterator++;
  }

  return retval;
}

//! Reset the read Iterator, so that getNextRead can recycle through list
void ReadManager::resetReadIterator()
{
  m_readIterator = m_readMap.begin();
}

//! Finds a read based on a readId
/*! 
 *  @return Pointer to ReadData of requested name or NULL if not found
 */
ReadData * ReadManager::find(const string & readId) const
{
  ReadData * retval = NULL;
  ReadMap::const_iterator i = m_readMap.find(readId);

  if (i != m_readMap.end())
  {
    retval = i->second;
  }

  return retval;
}

//! Returns the number of reads stored in the map
int ReadManager::size() const
{
  return m_readMap.size();
}

//! Removes a read from the map, but does not delete the ReadData
void ReadManager::removeRead(const string & id)
{
  m_readMap.erase(id);
}

//! Inserts a read into the read map
/*! 
 *  @throws Exception If there is already a read with id in the map
 */
void ReadManager::insertRead(ReadData * rd)
{
  const string & id = rd->getId();

  if (find(id))
  {
    throw Exception(id + " is already in readMap", 
                        "ReadManager::insertRead");
  }

  m_readMap[id] = rd;
}

//! Removes a read from the map under one id, and inserts it as another
/*! 
 *  @see removeRead
 *  @see insertRead
 */
void ReadManager::resetReadId(ReadData * rd, const string & id)
{
  const string & origId = rd->getId();

  if (find(origId))
  {
    removeRead(origId);
    rd->resetId(id);
    insertRead(rd);
  }
}

//! Prints the ReadCollection in XML with all of the reads
void ReadManager::printXML(ostream & out, const string & prefix) const
{
  out << prefix << "<ReadCollection>" << endl;

  ReadMap::const_iterator i;
  for (i =  m_readMap.begin();
       i != m_readMap.end();
       i++)
  {
    i->second->printXML(out, prefix + "  ");
  }

  out << prefix << "</ReadCollection>" << endl;
}

} // namespace
